﻿using System;
using System.Collections.Generic;
using System.Collections;
using System.Collections.Specialized;
using System.Linq;
using System.Text;
using Multimap;

namespace BayesNetwork.Classes
{
    public class BayesNet
    {
        public Dictionary<string, BayesNode> Nodes { get; set; }
        public Dictionary<string, BayesNode> PreviousNodes { get; set; }
        public Multimap<int, string> nodeLevels { get; set; }
        public Multimap<string, string> ME_nodes { get; set; }
        public Multimap<string, BayesEdge> Edges {get; set;} // WEIRD: You can have duplicate edges.

        public BayesNet()
        {
            Nodes = new Dictionary<string, BayesNode>();
            Edges = new Multimap<string,BayesEdge>();
            ME_nodes = new Multimap<string, string>();
            nodeLevels = new Multimap<int, string>();
        }

        #region addNode functions
        //standard node
        public void addNode(string name, double probability, int level)
        {
            Nodes.Add(name, new BayesNode(probability) { Name = name });
            nodeLevels.Add(level, name);
        }

        //copying nodes
        public void addNode(string nodeName, BayesNode node, int level)
        {
            Nodes.Add(nodeName, node);
            nodeLevels.Add(level, nodeName);
        }

        //Nodes that are not effected by influence.
        public void addNode(string name, bool notInfluenced, bool? observedValue, int level)
        {
            Nodes.Add(name, new BayesNode(0.0) { Name = name, ManuallySet = notInfluenced, ObservedValue = observedValue });
            nodeLevels.Add(level, name);
        }

        public void addNode(string me_catName, string[] nodeNames, double[] nodeProbabilities, int level)
        {
            if(nodeNames.Length != nodeProbabilities.Length)
                throw new Exception("the arrays must be the same length... one name, one probability.");

            for (int i = 0; i < nodeNames.Length; i++)
            {
                Nodes.Add(nodeNames[i], new BayesNode(nodeProbabilities[i]) { Name = nodeNames[i], me_cat = me_catName });
                ME_nodes.Add(me_catName, nodeNames[i]);
                nodeLevels.Add(level, nodeNames[i]);
            }
            normalizeMEnodes(me_catName);
        }
        public void addNode(string me_catName, string[] nodeNames, int level)
        {
            double[] d = new double[nodeNames.Length];
            for(int i=0; i<nodeNames.Length; i++)
                d[i] = 1.0/nodeNames.Length;

            addNode(me_catName, nodeNames, d, level); 
        }
        public void addNode(string me_catName, string nodeName, double nodeProbability, int level)
        {
            //Try to avoid using this function. Use the one above and declare all M.E. nodes at once.
            Nodes.Add(nodeName, new BayesNode(nodeProbability) { Name = nodeName });
            ME_nodes.Add(me_catName, nodeName);
            normalizeMEnodes(me_catName);
            nodeLevels.Add(level, nodeName);
        }
        #endregion

        #region addEdge Functions
        //Their is currently no normalizing of mutually exclusive edges... beware!
        //regular addEdge function
        public void addEdge(string nodeInfluencer, string nodeInfluenced, double influenceTrue, double influenceFalse)
        {
            // Ensure the value is within the acceptable range.
            if (0 > influenceTrue || influenceTrue > 2 ||
                0 > influenceFalse || influenceFalse > 2)
            {
                // Throw exception
                return;
            }
            BayesEdge newEdge = new BayesEdge(nodeInfluencer, nodeInfluenced, influenceTrue, influenceFalse);
            Edges.Add(nodeInfluenced, newEdge);
        }
        //addEdge with default opposite value for false probability.
        public void addEdge(string nodeInfluencer, string nodeInfluenced, double influenceTrue)
        {
            addEdge(nodeInfluencer, nodeInfluenced, influenceTrue, 2- influenceTrue);
        }
        /// <summary>
        /// all negative influencers are set to one in this case... I know it's inconsistant, however it is convinient... I'll fix it later.
        /// </summary>
        /// <param name="nodeInfluencer"></param>
        /// <param name="nodeInfluenced"></param>
        /// <param name="probabilities"></param>
        /// <param name="negateInfluence">negate the influence for the negative rather than a non-influencing 1.0.</param>
        public void addEdge(string nodeInfluencer, string[] nodeInfluenced, double[] probabilities, bool pastInfluence, bool negateInfluence)
        {
            if(nodeInfluenced.Length != probabilities.Length)
                throw new Exception("the passed in arrays, need to be the same lengths >>;;");

            double sum=0;
            foreach(double p in probabilities)
                sum += p;

            //Now with normalizing! (the normal for edges would be 1.0 across the board, therefore normalizing involves also multiplying the value by length)
            for (int i = 0; i < nodeInfluenced.Length; i++)
            {
                double newProbability = (probabilities[i] / sum) * probabilities.Length;
                BayesEdge newEdge = new BayesEdge(nodeInfluencer, nodeInfluenced[i], newProbability,
                    (negateInfluence) ? 2 - newProbability : 1.0) { PastInfluence = pastInfluence };
                Edges.Add(nodeInfluenced[i], newEdge);
            }
        }
        public void addEdge(string nodeInfluencer, string[] nodeInfluenced, double[] probabilities, bool pastInfluence)
        { addEdge(nodeInfluencer, nodeInfluenced, probabilities, pastInfluence, false); }
        public void addEdge(string nodeInfluencer, string[] nodeInfluenced, double[] probabilities)
        { addEdge(nodeInfluencer, nodeInfluenced, probabilities, false, false); }
        #endregion

        public double infer(string nodeName)
        {
            /*
             * Inferance is finding out P(X | p(X) where X is the current node and p(X) is all the parents of that node
             * 
             * To simplify inputing information each parents has a weight on how it effects its children.
             * This makes it so that for every parent we do not need to enter 2^|p| entry because
             * they can be calculated using the weighted function
             */

            BayesNode node = Nodes[nodeName];

            //new case! If the node is "NotInfluenced" as in nothing influences it, 
            //then return either 1 for true or 0 for false, to maintain the current value.
            if (node.ManuallySet)
                if (node.ObservedValue == null)
                    throw new Exception("All nodes that are not influenced, must have an observed value");
                else 
                    return (node.ObservedValue == true) ?1.0 :0.0 ;
            
            if(!Edges.ContainsKey(nodeName))
                return node.Probability;

            HashSet<BayesEdge> edges = Edges[nodeName];

            #region Probability formula Description
            /* 
             * Original formula:
             * P = P + (~P*(I-1))
             * Where P = Probability and I = Influence
             * 
             * We need to order of operation to have no effect. i.e. 1.2 then 0.8 needs to give the same result
             * as 0.8 then 1.2.
             * To fix this we let I be the average of all Influences
             * 
             * Now it was possible to come to a value less then 0. i.e. P = 0.2 and I = 0.2
             * To fix this we check if I is less then 1 ( a negative influnce) and we use the below
             * P = P + (P*(I-1)) 
             */
            #endregion

            double I = new double();
            int countI = new int();
            double P = node.Probability;

            foreach (BayesEdge edge in Edges.GetValues(node.Name, true))
            {
                
                BayesNode Influencer;
                if(edge.PastInfluence)
                    Influencer = PreviousNodes[edge.Influencer];
                else
                    Influencer = Nodes[edge.Influencer];

                bool? InfluencerValue = Influencer.ObservedValue;
                /* 
                 * If InfluencerValue is not set, do nothing
                 * If it is set and true add influence to the sum
                 * If it is set and false add the inverse of the influence to the sum 
                 */
                if (InfluencerValue.HasValue)
                    if (InfluencerValue.Value)
                    {
                        I += edge.Influence;
                        ++countI;
                    }
                    else
                    {
                        I += edge.NegativeInfluence;
                        ++countI;
                    }
            }
            if (I == 0)
                return P;

            I = I / countI;

            double FinalProbability = new double();
            if (I > 1)
                FinalProbability = P + ((1 - P) * (I - 1));
            else
                FinalProbability = P + (P * (I - 1));
            /*
             * We now have the probability of X given its parents.
             * We return this as the chance of success.
             */

            return FinalProbability;
        }

        public double conditional(string nodeName)
        {
            /*
             * Given the node X we can determin properties of a parent of X using 
             * conditional probability formula
             */
            BayesNode node = Nodes[nodeName];
            if (!Edges.ContainsKey(nodeName))
                return node.Probability;

            HashSet<BayesEdge> edges = Edges[nodeName];
            return node.Probability;
        }

        /* 
         * To overload the array operator. 
         * Example: You can use "bayesNet['nodeName']" to get a node.
         */
        public BayesNode this[string nodeName]
        {
            get { return Nodes[nodeName]; }
            set { Nodes[nodeName] = value; }
        }

        // Analysis Tools
        public HashSet<BayesNode> getNodesInfluencers(string nodeName)
        {
            HashSet<BayesNode> influencersNodes = new HashSet<BayesNode>();

            // If node doesn't exist, do nothing and return.
            if (Nodes[nodeName] == null)
                return influencersNodes;

            foreach (BayesEdge e in Edges.GetValues(nodeName, true))
                    influencersNodes.Add(Nodes[e.Influencer]);

            return influencersNodes;
        }

        public HashSet<BayesNode> getNodesInfluencees(string nodeName)
        {
            HashSet<BayesNode> influencedNodes = new HashSet<BayesNode>();

            // If node doesn't exist, do nothing and return.
            if (Nodes[nodeName] == null)
                return influencedNodes;

            foreach (KeyValuePair<string, HashSet<BayesEdge>> p in Edges)
                foreach (BayesEdge n in p.Value)
                    if (n.Influencer == nodeName)
                        influencedNodes.Add(Nodes[n.Influenced]);

            return influencedNodes;
        }

        public Boolean updateFull()
        {
            /*
             * This function sets the value of everything in the network to be either true or false
             * to be used by the environment.
             */
            Random randnum = new Random();
            double next = new double();
            double p = new double();

            PreviousNodes = new Dictionary<string,BayesNode>(Nodes);

            //fill in the details for mutually exclusive nodes.
            //For all categories
            //This is terrible effeciency refactor later... (Dec 9th 2009)
            //it is also very bad to use hard coded integers... but whatever
            for (int i = 0; i <= 3; i++)
            {
                foreach (string t in ME_nodes.Keys)
                {
                    next = randnum.NextDouble();
                    p = 0.0;
                    //for all nodes in a specific category.
                    bool observedFound = false;
                    foreach (string a in ME_nodes.GetValues(t))
                    {
                        if(!nodeLevels[i].Contains(a))
                            break; //this is lazy... this check should be before the first forloop.

                        p += this.infer(a);
                        if (p < next || observedFound)
                            this[a].ObservedValue = false;
                        else
                            this[a].ObservedValue = observedFound = true;
                    }
                }

                //fill in the details for non-mutually exclusive nodes
                foreach (KeyValuePair<string, BayesNode> b in Nodes)
                {
                    if (b.Value.me_cat == null && nodeLevels[i].Contains(b.Key))
                    {
                        p = this.infer(b.Key);
                        next = randnum.NextDouble();
                        if (p < next)
                            this[b.Key].ObservedValue = false;
                        else
                            this[b.Key].ObservedValue = true;
                    }
                }
            }

            return true;
        }

        public void normalizeMEnodes(string me_cat)
        {
            if (!ME_nodes.Keys.Contains(me_cat))
                throw new Exception("You cannot normalize what is not there... Currently that makes no sense.");

            HashSet<string> nodesInCategory = ME_nodes.GetValues(me_cat);
            
            //get the probability sum.
            double sum = 0;
            foreach (string nodeName in nodesInCategory)
                sum += Nodes[nodeName].Probability;
            //Now normalize!
            foreach (string nodeName in nodesInCategory)
                Nodes[nodeName].Probability /= sum;
        }
    }
}
